//
// This file is part of GENFERM.
// 
// Copyright 2005, 2022 Dario Alejandro Alpern
//
// GENFERM is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GENFERM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GENFERM.  If not, see <http://www.gnu.org/licenses/>.
//
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdint.h>
#include <stdbool.h>
#include <assert.h>
#ifdef _WIN32
#define FORMAT_INT64  "lld"
#else
#define FORMAT_INT64  "ld"
#endif
#define  MAX_UNSIGNED_LONG     0xFFFFFFFFUL
#define  LONGLONG_SIGN_MASK    0x8000000000000000ULL
#define  DOUBLE_SIZE_LONG      ((double)65536L*(double)65536L)
#define  DOUBLE_SIZE_LONGLONG  (DOUBLE_SIZE_LONG*DOUBLE_SIZE_LONG)
#define  MAX_NUMBER_LENGTH     10000
#define  MAX_PRIME             100000
#define  SIEVE_LENGTH          200000
int64_t Power3[MAX_NUMBER_LENGTH];
int64_t Base[MAX_NUMBER_LENGTH];
int64_t base[MAX_NUMBER_LENGTH];
int64_t BigTemp[MAX_NUMBER_LENGTH];
int64_t BigTemp2[MAX_NUMBER_LENGTH];
int64_t power[MAX_NUMBER_LENGTH];
int64_t Modulus[MAX_NUMBER_LENGTH];
uint32_t MontgomeryMultN;
int64_t MontgomeryMultR1[MAX_NUMBER_LENGTH];
double dN;
int MaxNumberLength;
int NumberLength;
int exponent;
int primeExp = 3;
int largestPower2PrimeExp;
int64_t minMultiplier;
int64_t maxMultiplier;
int64_t multiplier;
FILE *fpOut;
int primes[MAX_PRIME/10];
int modprimes[MAX_PRIME/10];
int deltamodprimes[MAX_PRIME/10];
char sieve[SIEVE_LENGTH];
int nbrPrimes;
void MontgomeryMult(const int64_t *Nbr1, const int64_t *Nbr2, int64_t *Prod);

int64_t AdjustModN(int64_t *Nbr)
{
  uint64_t TrialQuotient;
  uint64_t Cy;
  int i;
  double dAux;

  dAux = (double)*(Nbr + NumberLength) * DOUBLE_SIZE_LONG +
    (double)*(Nbr + NumberLength - 1);
  if (NumberLength >= 2)
  {
    dAux += (double)*(Nbr + NumberLength - 2) / DOUBLE_SIZE_LONG;
  }
  TrialQuotient = (uint64_t)(dAux / dN) + 3;
  if (TrialQuotient > MAX_UNSIGNED_LONG)
  {
    Cy = 0;
    for (i=0; i<NumberLength; i++)
    {
      Cy = *(Nbr+i+1)-(TrialQuotient>>32)*Modulus[i] - Cy;
      *(Nbr+i+1) = Cy & MAX_UNSIGNED_LONG;
      Cy = (MAX_UNSIGNED_LONG - Cy) >> 32;
    }
    TrialQuotient &= MAX_UNSIGNED_LONG;
  }
  Cy = 0;
  for (i=0; i<NumberLength; i++)
  {
    Cy = *(Nbr+i) - TrialQuotient*Modulus[i] - Cy;
    *(Nbr+i) = Cy & MAX_UNSIGNED_LONG;
    Cy = (MAX_UNSIGNED_LONG - Cy) >> 32;
  }
  *(Nbr+NumberLength) -= Cy;
  while ((*(Nbr+NumberLength) & MAX_UNSIGNED_LONG) != 0)
  {
    Cy=0;
    for (i=0; i<NumberLength; i++)
    {
      Cy += *(Nbr+i)+Modulus[i];
      *(Nbr+i) = Cy & MAX_UNSIGNED_LONG;
      Cy >>= 32;
    }
    *(Nbr+NumberLength) += Cy;
    TrialQuotient--;
  }
  return TrialQuotient;
}

void MultBigNbrModN(const int64_t *Nbr1, const int64_t *Nbr2, int64_t *Prod)
{
  int i;
  int j;
  uint64_t Pr;
  uint64_t Nbr;

  dN = (double)Modulus[NumberLength-1];
  dN += (double)Modulus[NumberLength-2]/DOUBLE_SIZE_LONG;
  dN += (double)Modulus[NumberLength-3]/DOUBLE_SIZE_LONGLONG;
  memset(Prod, 0, NumberLength*sizeof(Prod[0]));
  i = NumberLength;
  do
  {
    i--;
    Nbr = *(Nbr1+i);
    j = NumberLength;
    do
    {
      *(Prod+j) = *(Prod+j-1);
      j--;
    } while (j > 0);
    *Prod = 0;
    Pr = 0;
    for (j = 0; j < NumberLength; j++)
    {
      Pr = (Pr >> 32) + Nbr * *(Nbr2+j) + *(Prod+j);
      *(Prod+j) = Pr & MAX_UNSIGNED_LONG;
    }
    *(Prod+j) += (Pr >> 32);
    AdjustModN(Prod);
  } while (i > 0);
}

void MultBigNbrByLong(const int64_t *Nbr1, int64_t Nbr2, int64_t *Prod)
{
  int64_t Pr;
  int j;

  Pr = 0;
  for (j = 0; j < NumberLength; j++)
  {
    Pr = (Pr >> 32) + Nbr2 * *(Nbr1+j);
    *(Prod+j) = Pr & MAX_UNSIGNED_LONG;
  }
  *(Prod+j) = (Pr >> 32);
}

void MultBigNbrByLongModN(const int64_t *Nbr1, int64_t Nbr2, int64_t *Prod)
{
  MultBigNbrByLong(Nbr1, Nbr2, Prod);
  AdjustModN(Prod);
}

void AddBigNbrModN(const int64_t *Nbr1, const int64_t *Nbr2, int64_t *Sum)
{
  int64_t Cy=0;
  int i;

  for (i=0; i<NumberLength; i++)
  {
    Cy += *(Nbr1+i) + *(Nbr2+i) - Modulus[i];
    *(Sum+i) = Cy & MAX_UNSIGNED_LONG;
    Cy >>= 32;
  }
  if (Cy < 0)
  {
    Cy = 0;
    for (i=0; i<NumberLength; i++)
    {
      Cy += *(Sum+i) + Modulus[i];
      *(Sum+i) = Cy & MAX_UNSIGNED_LONG;
      Cy >>= 32;
    }
  }
}

void SubtractBigNbrModN(const int64_t *Nbr1, const int64_t *Nbr2, int64_t *Result)
{
  int64_t Cy=0;
  int i;

  for (i=0; i<NumberLength; i++)
  {
    Cy += *(Nbr1+i) - *(Nbr2+i);
    *(Result+i) = Cy & MAX_UNSIGNED_LONG;
    Cy >>= 32;
  }
  if (Cy < 0)
  {
    Cy = 0;
    for (i=0; i<NumberLength; i++)
    {
      Cy += *(Result+i) + Modulus[i];
      *(Result+i) = Cy & MAX_UNSIGNED_LONG;
      Cy >>= 32;
    }
  }
}

void AddBigNbr(const int64_t *Nbr1, const int64_t *Nbr2, int64_t *Sum)
{
  int64_t Cy=0;

  for (int i=0; i<NumberLength; i++)
  {
    Cy += *(Nbr1+i) + *(Nbr2+i);
    *(Sum+i) = Cy & MAX_UNSIGNED_LONG;
    Cy >>= 32;
  }
}

void SubtractBigNbr(const int64_t *Nbr1, const int64_t *Nbr2, int64_t *Result)
{
  int64_t Cy=0;

  for (int i=0; i<NumberLength; i++)
  {
    Cy += *(Nbr1+i) - *(Nbr2+i);
    *(Result+i) = Cy & MAX_UNSIGNED_LONG;
    Cy >>= 32;
  }
}

int64_t RemDivBigNbrByLong(const int64_t *Dividend, int64_t Divisor)
{
  int i;
  int64_t Divid;
  int64_t Rem = 0;

  if (Divisor < 0)
  {                            // If divisor is negative...
    Divisor = -Divisor;        // Convert divisor to positive.
  }
  i = NumberLength - 1;
  if (*(Dividend+i) >= 0x80000000L)
  {                            // If dividend is negative...
    Rem = Divisor - 1;
  }
  for ( ; i >= 0; i--)
  {
    Divid = *(Dividend+i) + (Rem << 32);
    Rem = Divid % Divisor;
  }
  return Rem;
}

void GetMontgomeryParms(void)
{
  unsigned long mod0;

  dN = (double)Modulus[NumberLength-1];
  if (NumberLength > 1)
  {
    dN += (double)Modulus[NumberLength - 2] / DOUBLE_SIZE_LONG;
  }
  if (NumberLength > 2)
  {
    dN += (double)Modulus[NumberLength - 3] / DOUBLE_SIZE_LONGLONG;
  }
  mod0 = (unsigned long)Modulus[0];
  MontgomeryMultN = 2 - mod0;
  MontgomeryMultN *= 2 - (MontgomeryMultN * mod0);
  MontgomeryMultN *= 2 - (MontgomeryMultN * mod0);
  MontgomeryMultN *= 2 - (MontgomeryMultN * mod0);
  MontgomeryMultN *= 2 - (MontgomeryMultN * mod0);
  MontgomeryMultN *= 2 - (MontgomeryMultN * mod0);
  memset(MontgomeryMultR1, 0, (NumberLength+1)*sizeof(MontgomeryMultR1[0]));
  MontgomeryMultR1[NumberLength] = 1;
  AdjustModN(MontgomeryMultR1);
}

void RaiseToPrimeExp(void)
{
  memcpy(BigTemp, power, NumberLength*sizeof(BigTemp[0]));
  for (int i = largestPower2PrimeExp; i > 0; i /= 2)
  {
    MontgomeryMult(BigTemp, BigTemp, BigTemp2);
    if (primeExp & i)
    {             // Square and multiply by power.
      MontgomeryMult(BigTemp2, power, BigTemp);
    }
    else
    {             // Square only.
      memcpy(BigTemp, BigTemp2, NumberLength*sizeof(BigTemp[0]));
    }
  }
  memcpy(power, BigTemp, NumberLength*sizeof(BigTemp[0]));
}

bool BigNbrIsZero(const int64_t *nbr)
{
  for (int i=0; i<NumberLength; i++)
  {
    if (*(nbr + i) != 0)
    {
      return false;            // Number is not zero.
    }
  }
  return true;                 // Number is zero.
}

void TestGenFerm(int expon)
{
  bool Plus1IsMultiple = false;
  bool Minus1IsMultiple = false;
  char info[300];
               // Perform 2^(primeExp^expon) mod Modulus
  GetMontgomeryParms();
  base[0] = 2;
  memset(base+1, 0, NumberLength*sizeof(base[0]));
  MultBigNbrModN(base, MontgomeryMultR1, power);
  for (int i=0; i<expon; i++)
  {
    RaiseToPrimeExp();
  }
  AddBigNbrModN(power, MontgomeryMultR1, BigTemp);
  if (BigNbrIsZero(BigTemp))
  {
    Plus1IsMultiple = true;
  }
  if (!memcmp(power, MontgomeryMultR1, NumberLength*sizeof(power[0])))
  {
    Minus1IsMultiple = true;
  }
  RaiseToPrimeExp();
  AddBigNbrModN(power, MontgomeryMultR1, BigTemp);
  if (BigNbrIsZero(BigTemp) && !Plus1IsMultiple)
  {
    sprintf(info, "(2^%d^%d + 1)/(2^%d^%d + 1) has a divisor %" FORMAT_INT64 " * %d^%d + 1\n",
            primeExp, expon+1, primeExp, expon, 2*multiplier, primeExp, expon+1);
    printf("%s", info);
    fprintf(fpOut, "%s", info);fflush(fpOut);
  }
  if (!memcmp(power, MontgomeryMultR1, NumberLength*sizeof(power[0])) &&
      !Minus1IsMultiple)
  {
    sprintf(info, "(2^%d^%d - 1)/(2^%d^%d - 1) has a divisor %" FORMAT_INT64 " * %d^%d + 1\n",
            primeExp, expon+1, primeExp, expon, 2*multiplier, primeExp, expon+1);
    printf("%s", info);
    fprintf(fpOut, "%s", info);fflush(fpOut);
  }
}

int64_t modInv(int NbrMod, int currentPrime)
{
  int64_t QQ;
  int64_t T1;
  int64_t T3;
  int64_t U1 = 1;
  int64_t U3 = NbrMod;
  int64_t V1 = 0;
  int64_t V3 = currentPrime;
  assert(currentPrime > 2);
  while (V3 != 0)
  {
    QQ = U3 / V3;
    T1 = U1 - V1 * QQ;
    T3 = U3 - V3 * QQ;
    U1 = V1;
    U3 = V3;
    V1 = T1;
    V3 = T3;
  }
  return (U1 + currentPrime) % currentPrime;
}

static void adjustNumberLength(void)
{
  for (NumberLength = MaxNumberLength; NumberLength >= 2; NumberLength--)
  {
    if ((Modulus[NumberLength - 1] != 0) ||
      (Modulus[NumberLength - 2] > MAX_UNSIGNED_LONG / 2))
    {
      break;
    }
  }
}

static void multiplierLoop(int64_t show)
{
  int i;
  int P;
  for (;;)
  {
    memset(sieve, 0, SIEVE_LENGTH);
    for (i = 0; i < nbrPrimes; i++)
    {
      P = primes[i];
      for (int j = modprimes[i]; j < SIEVE_LENGTH; j += P)
      {
        sieve[j] = 1;          // Composite
      }
      modprimes[i] = (modprimes[i] + deltamodprimes[i]) % P;
    }

    for (i = 0; i < SIEVE_LENGTH; i++)
    {
      if (sieve[i] == 0)
      {
        // Possibly prime.
        memcpy(Modulus, Base, NumberLength * sizeof(Base[0]));
        adjustNumberLength();
        TestGenFerm(exponent);
        NumberLength = MaxNumberLength;
      }
      multiplier++;
      if ((show == 0) || (multiplier % show == 0))
      {
        printf("Exponent = %d, Multiplier = %-16" FORMAT_INT64 "               \r", exponent, multiplier * 2);
#ifndef _WIN32
        fflush(stdout);
#endif
      }
      if (multiplier > maxMultiplier)
      {
        exponent++;
        return;
      }
      AddBigNbr(Base, Power3, Base);
    }
  }
}

static bool isPrime(int nbr)
{
  for (int divisor = 3; (divisor * divisor) <= nbr; divisor += 2)
  {                   // Check if primeExp is prime.
    if ((nbr % divisor) == 0)
    {                 // Composite.
      return false;
    }
  }
  return true;
}

static int nextPrime(int nbr)
{
  int candidate = nbr;
  do
  {
    candidate += 2;
  } while (!isPrime(candidate));
  return candidate;
}

static int exponentLoop(void)
{
  int64_t multiplierLow;
  int64_t multiplierHigh;
  int64_t Cy;
  int64_t show;
  int64_t valuesievezero;
  int i;
  int prime;
  int index;
  int delta;
  for (;;)
  {     // For each exponent...
    multiplier = minMultiplier;
    // Compute 2*primeExp times the power of primeExp.
    MaxNumberLength = (int)((double)exponent * log(primeExp) / (32 * log(2)) + 3);
    if (MaxNumberLength > MAX_NUMBER_LENGTH - 5)
    {
      printf("Exponent too large.\n");
      return 2;
    }
    fpOut = fopen("genferm.dat", "a");
    NumberLength = MaxNumberLength;
    Power3[0] = 2 * primeExp;
    memset(Power3 + 1, 0, (NumberLength - 1) * sizeof(Power3[0]));
    for (i = 0; i < exponent; i++)
    {
      MultBigNbrByLong(Power3, primeExp, Power3);
    }
    // Compute first base.
    multiplierLow = multiplier & MAX_UNSIGNED_LONG;
    multiplierHigh = (multiplier >> 32) & MAX_UNSIGNED_LONG;
    MultBigNbrByLong(Power3, multiplierLow, BigTemp);
    MultBigNbrByLong(Power3, multiplierHigh, BigTemp2 + 1);
    BigTemp2[0] = 0;
    AddBigNbr(BigTemp, BigTemp2, Base);
    Cy = 1;
    for (i = 0; i < NumberLength; i++)
    {
      Cy += Base[i];
      Base[i] = Cy & MAX_UNSIGNED_LONG;
      Cy >>= 32;
    }
    show = (int64_t)(pow(10, floor(9.6 - 2 * log(exponent) / log(10)))) / 2;

    // Compute primes.
    prime = 5;
    index = 0;
    while (prime < MAX_PRIME)
    {
      primes[index] = prime; /* Store prime */
      valuesievezero = RemDivBigNbrByLong(Base, prime);
      delta = (int)RemDivBigNbrByLong(Power3, prime);
      modprimes[index] = (int)(prime - ((modInv(delta, prime) * valuesievezero) % prime)) % prime;
      deltamodprimes[index] = (prime - SIEVE_LENGTH % prime) % prime;
      // Compute next prime.
      prime = nextPrime(prime);
      index++;
    }
    nbrPrimes = index;
    multiplierLoop(show);
  }
}

static int parseCommandLine(int argc, char* argv[])
{
  for (int i = 1; i < argc; i++)
  {
    if (strlen(argv[i]) > 3 && !memcmp(argv[i], "-p:", 3))
    {
      primeExp = atoi(&argv[i][3]);
      if ((primeExp < 2) || ((primeExp > 2) && (primeExp % 2 == 0)))
      {
        printf("-p parameter must be prime.\n");
        return 1;
      }
      if (!isPrime(primeExp))
      {                 // Composite.
        printf("-p parameter must be prime.\n");
        return 1;
      }
    }
    else if ((strlen(argv[i]) > 3) &&
      (memcmp(argv[i], "-e:", 3) == 0))
    {
      exponent = atoi(&argv[i][3]);
    }
    else if (strlen(argv[i]) > 5 &&
      (memcmp(argv[i], "-min:", 5) == 0))
    {
#ifdef _WIN32
      minMultiplier = _atoi64(&argv[i][5]) / 2;
#else
      minMultiplier = atoll(&argv[i][5]) / 2;
#endif
    }
    else if (strlen(argv[i]) > 5 &&
      (memcmp(argv[i], "-max:", 5) == 0))
    {
#ifdef _WIN32
      maxMultiplier = _atoi64(&argv[i][5]) / 2;
#else
      maxMultiplier = atoll(&argv[i][5]) / 2;
#endif
    }
  }
  return 0;
}

int main(int argc, char* argv[])
{
  maxMultiplier = 1000000000000000000LL;
  int rc = parseCommandLine(argc, argv);
  if (rc != 0)
  {
    return rc;
  }
  if ((argc < 2) || (exponent == 0))
  {
    printf("GENFERM computes factors of (2^p^(e+1)-1)/(2^p^e-1) and (2^p^(e+1)+1)/(2^p^e+1).\n");
    printf("Written by Dario Alejandro Alpern on December 2022.\n\n");
    printf("Command line: genferm [-p:prime] -e:exponent\n");
    printf("                      [-min:first_multiplier] [-max:maximum_multiplier]\n\n");
    printf("Defaults: prime = 3, min = 0, max = 10^18\n");
    return 1;
  }
  largestPower2PrimeExp = 0x40000000;
  while (largestPower2PrimeExp > 0)
  {
    if (largestPower2PrimeExp & primeExp)
    {
      largestPower2PrimeExp /= 2;
      break;
    }
    largestPower2PrimeExp /= 2;
  }
  return exponentLoop();
}
